LISP Model
==================================

Lisp model works well; sets the order of operations.  Relatively clear, although somewhat nasty also.  Will run into trouble for array values?  Not if already in a hash.  Not natural for JSON.

    - - set
      - 0
      - - build
        - task
        - load
        
    - - set
      - 1
      - - build
        - task
        - dump
    - - build
      - join
      - - ref
        - 0
      - - ref
        - 1
        
    - - enq
      - - ref
        - 0
      - - data
        - 0
        - 1
        - 2
        
        
    0:
      0: set
      1: 0
      2:
        0: build
        1: task
        2: load
    1:
      0: set
      1: 0
      2:
        0: build
        1:
          _type:
          _class
    2:
      0: build
      1: join
      2:
        0: ref
        1: 0
      3:
        0: ref
        1: 1
    3:
      - enq
      - - ref
        - 0
      - - data
        - 0
        - 1
        - 2

Method Model
==================================

Method model is relatively clear, fewer issues; strongly tied to schema.  Unnatural for JSON.  Not clear how to resolve order.

    _id:
    _rev:
    schema:
    - - 0
      - build
      - task
      - load
    - - 1
      - build
      - task
      - dump
    - - 
      - build
      - join
      - join
      - - 0
      - - 1
    - - 
      - enq
      - 0
      - - 0
        - 1
        - 2
        
    class Schema
      def build(key, spec)
      end
  
      def enq(key, data)
      end
    end

Entries map like:

    - - var
      - method
      - arg
      - arg
      - arg
    
    var = schema.method(*args)

For example:

    - - var
      - build
      - _class:
        _type:
        config: 

    var = schema.build(spec)
    
Current Model
========================

Clear, inhomogeneous.  Referencing not that clear; non-tasks cannot be referenced and join structure is fixed.  Respects the basic differences of these objects and preserves order where necessary.

    tasks:
      key: spec
    joins:
    - [[inputs], [outputs], spec]
    middleware:
    - spec
    queue:
    - [key, data]

Type Model
=======================

More homogeneous; forces queue to be defined as a type, which is lousy.  Allows order when desired, but remains confusing for merge; what if it is an array in one and a hash in another?

    task:
      key: spec
    join:
    - spec
    middleware:
    - spec
    queue:
    - spec

Table Model
=======================

Requires use of _refs in the queue spec.  Nice that anything can be referenced, but no sure order.  Also requires queue to be defined as a resource type.  Very extensible; amenable to JSON, homogeonous.

    name: 
      - task
      - load
    key: 
      - task
      - dump
    2: 
      - join
      - join
      - - name
      - - key
    3:
      - queue
      - enq
      - name
      - a
      - b
      - c
      
List Model
=======================

Kind of nice in that it clearly delineates the workflow from the queue, which are distinct.  Obvious method for merging schemas, referencing external resources, etc.  No mixup between resources like data (referenced by inputs in the queue, for instance) and workflow elements.  Control still in the hands of the resource for references.  Method to NOT persist workflow resources, but just build them.  Obvious way to rename.  Leave it to the user to ensure the workflow is correctly ordered.  Structured and consistent.  Should parse easily.  Can detect name collisions.  Can detect disorder in spec sequence.  Can detect missing refs, with name and type (at least to the correct array index, or the full name in JSON).

From JSON:

    _id:
    _rev:
    workflow:
    -                     # inline resource schema, amenable to rename
      _id:
      _rev:
      _var: 0
      _type: task
      _class: load
    -                     # _var caches resource, optional
      _id:
      _rev:
      _var: 1
      _type: task
      _class: dump
    -                     # refs are inferred as array values
      _id:
      _rev:
      _type: join
      _class: join
      inputs: [0]
      outputs: [1]
    - uri                 # opens, loads as JSON...
                          # any fields (_var, etc)
                          # are accepted directly
                        
    - _id:                # arbitrary resource spec
      _rev:
      _type: type
      _class: arbitrary   
      _refs: [refs]       # || all array values
      _method: build      # || build
      _spec:              # || self
        refs: [0]
        key: value
    queue:
    -                     # relative names, allows queues to be
      _id:                # distinct resources, and resued
      _rev:
      task: 0             # references a workflow variable
      data: [a, b, c]     # may be uri for reference to JSON data, optional
    - uri

From parse:

    workflow:
    - - 0       # -- load
      - task
      - load
    - - 1       # -- dump
      - task
      - dump
    - -         # --[0][1]
      - join
      - join
      - - 0
      - - 1
    - -         # --/debugger
      - middleware
      - debugger
    - - 2       # --. type arbitrary -[0]- arg
      - type
      - arbitrary
      - - 0
      - arg
    queue:      # --@ 0 a b c
    - - 0
      - a
      - b
      - c

Entries map like:

    - 
      _var:
      _type:
      _class:
      config:
  
    # schema
    var = _class.build({:config => {}}, app)
  
    -
      task: 0
      inputs: [a, b, c]
    
    # queue
    app.enq(cache[0], inputs)

This is additionally nice because it can be turned into a task, given it's super-simple mapping.

    class Schema
      def build(spec, app, env)
        case spec
        when Array
          _var, _type, _class, *args = spec
      
          resource = env[_type][_class].parse!(args, app)
          cache[_var] = resource if _var
    
        when Hash
          _var = spec['_var']
          _type = spec['_type']
          _class = spec['_class']
    
          resource = env[_type][_class].build(spec, app)
          cache[_var] = resource if _var
    
        when String
          build fetch(spec), app, env
    
        else
          raise "error"
        end
      end
  
      def enque(spec)
        case spec
        when Array
          task, inputs = spec
          cache[task].enq(*inputs)
      
        when Hash
          enque [spec['task'], spec['inputs']]
      
        when String
          enque fetch(spec)
      
        else
          raise "error"
        end
      end
    end

Mixed method as data Model
=======================

Wow, what a lousy idea...

    -
      build:
        _type:
        _class:
      var: 0
    -
      build:
        _type:
        _class:
      var: 1
    -
      build:
        _type: join
        _class: join
        inputs:
        - 0
        outputs:
        - 1
    -
      enq:
        tasks:
        - 0
        data:
        - a
        - b
        - c


Breakdown
========================
This is tough because I'm genuinely asking for some of the niceties of a programming language.  Context mostly, but order as well.  I can get context, variables, using a hash, but not order.